from rask.base import Base
from tornado.web import asynchronous,RequestHandler
from tornado.websocket import WebSocketHandler

__all__ = [
    'asynchronous'
    'Handler',
    'WSHandler'
]

class Handler(RequestHandler,Base):
    def set_default_headers(self):
        self.set_header('Server','Viking Makt HTTP Server')

class WSHandler(WebSocketHandler,Base):
    def check_origin(self,origin):
        return True



# -*- coding: utf-8 -*-
from base64 import b64decode,b64encode
from rask.base import Base
from types import BooleanType,DictType,FloatType,IntType,ListType,LongType,NoneType,StringType,TupleType,UnicodeType

__all__ = ['UTCode']

class UTCode(Base):
    @property
    def decode_type(self):
        try:
            assert self.__decode_types
        except (AssertionError,AttributeError):
            self.__decode_types = {
                'b':self.decoder_bool,
                'd':self.decoder_dict,
                'f':self.decoder_float,
                'i':self.decoder_int,
                'l':self.decoder_list,
                'n':self.decoder_none,
                's':self.decoder_string,
                'u':self.decoder_unicode
            }
        except:
            raise
        return self.__decode_types
    
    @property
    def encode_types(self):
        try:
            assert self.__encode_types
        except (AssertionError,AttributeError):
            self.__encode_types = {
                BooleanType:self.encoder_bool,
                DictType:self.encoder_dict,
                FloatType:self.encoder_float,
                IntType:self.encoder_int,
                ListType:self.encoder_list,
                LongType:self.encoder_int,
                NoneType:self.encoder_none,
                StringType:self.encoder_string,
                TupleType:self.encoder_list,
                UnicodeType:self.encoder_unicode
            }
        except:
            raise
        return self.__encode_types

    def decode(self,_,future,result=None):
        try:
            assert _[:3] == 'ut:'
        except (AssertionError,TypeError):
            future.set_result(None)
            self.log.error('Invalid UTCode')
            return False
        except:
            raise
        
        def on_future(r):
            future.set_result(r.result()[0])
            return True

        self.ioengine.loop.add_callback(
            self.decode_type[_[3]],
            _=_[3:],
            f=self.ioengine.future(on_future)
        )
        return True

    def decoder_bool(self,_,f):
        f.set_result(([False,True][int(_[2])],_[3:]))
        return True

    def decoder_dict(self,_,f,r=None):
        try:
            assert _[:2] == 'd:'
        except:
            raise
        else:
            def on_end(result):
                f.set_result(result.result())
                return True

            self.ioengine.loop.add_callback(
                self.decoder_dict_consumer,
                _=_[2:],
                f=self.ioengine.future(on_end),
                r={}
            )
        return True

    def decoder_dict_consumer(self,_,f,r=None):
        try:
            assert _[0] == 'e'
        except AssertionError:
            k,_ = self.decoder_dict_key(_)

            def on_future(result):
                r[k] = result.result()[0]

                self.ioengine.loop.add_callback(
                    self.decoder_dict_consumer,
                    _=result.result()[1],
                    f=f,
                    r=r
                )
                return True
            
            self.ioengine.loop.add_callback(
                self.decode_type[_[0]],
                _=_,
                f=self.ioengine.future(on_future)
            )
        except:
            raise
        else:
            f.set_result((r,_[1:]))
        return True

    def decoder_dict_key(self,_):
        c = _.index(':')
        l = int(_[1:c])
        return (_[c+1:l+c+1],_[l+c+1:])

    def decoder_float(self,_,f):
        f.set_result(
            (
                float(_[_.index(':')+1:_.index('z')]),
                _[_.index('z')+1:]
            )
        )
        return True
    
    def decoder_int(self,_,f):
        f.set_result(
            (
                int(_[_.index(':')+1:_.index('e')]),
                _[_.index('e')+1:]
            )
        )
        return True

    def decoder_list(self,_,f,r=None):
        try:
            assert _[:2] == 'l:'
        except:
            raise
        else:
            def on_end(result):
                f.set_result(result.result())
                return True
            
            self.ioengine.loop.add_callback(
                self.decoder_list_consumer,
                _=_[2:],
                f=self.ioengine.future(on_end),
                r=[]
            )
        return True

    def decoder_list_consumer(self,_,f,r=None):
        try:
            assert _[0] == 'e'
        except AssertionError:
            def on_future(result):
                r.append(result.result()[0])

                self.ioengine.loop.add_callback(
                    self.decoder_list_consumer,
                    _=result.result()[1],
                    f=f,
                    r=r
                )
                return True

            self.ioengine.loop.add_callback(
                self.decode_type[_[0]],
                _=_,
                f=self.ioengine.future(on_future)
            )
        except:
            raise
        else:
            f.set_result((r,_[1:]))
        return True    

    def decoder_none(self,_,f):
        f.set_result((None,_[3:]))
        return True
    
    def decoder_string(self,_,f):
        c = _.index(':')
        l = int(_[1:c])
        f.set_result((_[c+1:l+c+1],_[l+c+1:]))
        return True

    def decoder_unicode(self,_,f):
        c = _.index(':')
        l = int(_[1:c])
        f.set_result((b64decode(_[c+1:l+c+1]).decode('utf-8'),_[l+c+1:]))
        return True
    
    def encode(self,_,future,result=None):
        def on_future(r):
            result = ['ut:']
            result.append(r.result())
            future.set_result(''.join(result))
            return True
        
        self.ioengine.loop.add_callback(
            self.encode_types[type(_)],
            _=_,
            f=self.ioengine.future(on_future)
        )
        return True

    def encoder_bool(self,_,f):
        f.set_result('b:%s' % [0,1][_])
        return True
    
    def encoder_dict(self,_,f,i=None,r=None):
        try:
            k = i.next()
        except AttributeError:
            self.ioengine.loop.add_callback(
                self.encoder_dict,
                _=_,
                f=f,
                i=iter(_),
                r=['d:']
            )
        except StopIteration:
            r.append('e')
            f.set_result(''.join(r))
            return True
        except:
            raise
        else:
            def on_future(result):
                r.append('k%s:%s' % (len(k),k))
                r.append(result.result())

                self.ioengine.loop.add_callback(
                    self.encoder_dict,
                    _=_,
                    f=f,
                    i=i,
                    r=r
                )
                return True

            self.ioengine.loop.add_callback(
                self.encode_types[type(_[k])],
                _=_[k],
                f=self.ioengine.future(on_future)
            )
        return None

    def encoder_float(self,_,f):
        f.set_result('f:%sz' % _)
        return True
    
    def encoder_list(self,_,f,i=None,r=None):
        try:
            k = i.next()
        except AttributeError:
            self.ioengine.loop.add_callback(
                self.encoder_list,
                _=_,
                f=f,
                i=iter(_),
                r=['l:']
            )
        except StopIteration:
            r.append('e')
            f.set_result(''.join(r))
            return True
        except:
            raise
        else:
            def on_future(result):
                r.append(result.result())

                self.ioengine.loop.add_callback(
                    self.encoder_list,
                    _=_,
                    f=f,
                    i=i,
                    r=r
                )
                return True
            
            self.ioengine.loop.add_callback(
                self.encode_types[type(k)],
                _=k,
                f=self.ioengine.future(on_future)
            )
        return None

    def encoder_int(self,_,f):
        f.set_result('i:%se' % _)
        return True

    def encoder_none(self,_,f):
        f.set_result('n:e')
        return True
    
    def encoder_string(self,_,f):
        f.set_result('s%s:%s' % (len(_),_))
        return True
    
    def encoder_unicode(self,_,f):
        b = b64encode(_.encode('utf-8'))
        f.set_result('u%s:%s' % (len(b),b))
        return True




__all__ = ['ack']

def ack(channel,method):
    def f(_):
        try:
            assert _ and _.result()
        except AssertionError:
            channel.basic_nack(method.delivery_tag)
            return False
        except AttributeError:
            pass
        except:
            raise
        
        channel.basic_ack(method.delivery_tag)
        return True
    return f



# -*- coding: utf-8 -*-
from simplejson import dumps,loads

__all__ = [
    'dictfy',
    'jsonify'
]

def dictfy(_):
    try:
        return loads(_)
    except:
        return False

def jsonify(_):
    try:
        return dumps(_)
    except:
        return False



from rask.base import Base
from rask.options import options

__all__ = ['Channel']

class Channel(Base):
    __channel = False

    @property
    def channel(self):
        return self.__channel

    @channel.setter
    def channel(self,_):
        try:
            assert _
        except AssertionError:
            self.__channel = False
        except:
            raise
        else:
            self.__channel = _
            self.active = True

    def __init__(self,connection,future=None,prefetch=None):
        try:
            assert future
        except AssertionError:
            pass
        except:
            raise
        else:
            self.promises.append(future)
        finally:
            self.ioengine.loop.add_callback(
                self.open,
                connection=connection,
                prefetch=prefetch
            )
            
    def open(self,connection,prefetch=None):
        def on_open(result):
            try:
                assert result
            except AssertionError:
                future.set_result(False)
            except:
                raise
            else:
                self.channel = result
                self.ioengine.loop.add_callback(self.qos_set,prefetch=prefetch)
            return True

        connection.channel(on_open)
        return True

    def qos_set(self,prefetch=None):
        def on_qos(_):
            return True
        
        try:
            assert prefetch
        except AssertionError:
            prefetch = options.rask['rmq']['channel']['prefetch']
        except:
            raise

        self.channel.basic_qos(callback=on_qos,prefetch_count=prefetch)
        return True



from rask.base import Base
from rask.options import define,options,parse_command_line
from rask.parser.json import dictfy
from rask.parser.utcode import UTCode
from tornado.web import Application

__all__ = ['Main']

class Main(Base):
    __http__ = None
    __http_routes__ = []
    __options__ = {
        'rmq':{
            'channel':{
                'prefetch':10
            }
        }
    }
    __settings_decoders = {
        'json':dictfy
    }

    def __init__(self):
        self.before()
        self.setup()
        self.ioengine.loop.add_callback(self.after)
        self.ioengine.loop.add_callback(self.http)
        self.ioengine.start()

    @property
    def services(self):
        try:
            assert self.__services
        except (AssertionError,AttributeError):
            self.__services = {}
        except:
            raise
        return self.__services

    @property
    def utcode(self):
        try:
            assert self.__utcode
        except (AssertionError,AttributeError):
            self.__utcode = UTCode()
        except:
            raise
        return self.__utcode
    
    def __settings_file_read(self,name):
        try:
            rfile = open(name,'r')
            rcontent = rfile.read()
            rfile.close()
        except IOError:
            self.log.error('IOError on %s open file, stopping IOEngine...' % name)
            self.ioengine.stop()
        except:
            raise
        return rcontent

    def __setup_autoreload(self):
        try:
            assert options.autoreload
        except AssertionError:
            pass
        except:
            raise
        else:
            self.log.info('autoreload: start')
            
            from tornado import autoreload
            autoreload.start()
        return True
    
    def after(self):
        pass

    def before(self):
        pass

    def http(self):
        try:
            assert self.__http_routes__
        except AssertionError:
            return False
        except:
            raise
        else:
            self.__http__ = Application(self.__http_routes__)
            self.ioengine.loop.add_callback(
                self.__http__.listen,
                port=options.http_port,
                xheaders=True
            )
            self.log.info('HTTP Server - %s' % options.http_port)
        return True

    def services_init(self):
        pass
    
    def setup(self):
        define('autoreload',default=False)
        define('http_port',default=8088)
        define('rask',default=self.__options__)
        define('raven_color',default=True)
        define('raven_color_fmt',default='%(log_color)s%(levelname)1.1s %(asctime)s %(name)s%(reset)s %(message)s')
        define('raven_no_color_fmt',default='%(levelname)1.1s %(asctime)s %(name)s %(message)s')

        parse_command_line()
        
        self.ioengine.loop.add_callback(self.__setup_autoreload)
        return True

    def settings_load(self,name,f_name,encode='json',future=None):
        try:
            define(
                name,
                default=self.__settings_decoders[encode](self.__settings_file_read(f_name))
            )
        except KeyError:
            self.log.info('Settings load, no encode (%s) found, stopping IOEngine...' % encode)
            self.ioengine.stop()
        except:
            raise
        else:
            self.log.info('Settings %s loaded...' % name)

        try:
            future.set_result(True)
        except:
            pass            
        return True



from logging import getLogger
from logging.config import dictConfig
from rask.options import define,options
from tornado.process import task_id

__all__ = ['Raven']

class Raven(object):
    __logger = None
    __name = None

    @property
    def __formatter__(self):
        try:
            assert options.raven_color
        except AssertionError:
            return {
                'format':options.raven_no_color_fmt
            }
        except:
            raise
        return {
            '()':'colorlog.ColoredFormatter',
            'format':options.raven_color_fmt
        }

    @property
    def logger(self):
        try:
            assert self.__logger
        except AssertionError:
            self.__config__()
            self.__logger = getLogger(self.__name)
        except:
            raise
        return self.__logger

    def __config__(self):
        dictConfig({
            'disable_existing_loggers':False,
            'formatters':{
                'rask':self.__formatter__
            },
            'handlers':{
                'default':{
                    'level':str(options.logging).upper(),
                    'class':'logging.StreamHandler',
                    "formatter": "rask"
                }
            },
            'loggers':{
                '': {
                    'handlers': ['default'],
                    'level':str(options.logging).upper(),
                    'propagate':True
                }
            },
            'version':1
        })
        return True

    def __init__(self,name='rask'):
        self.__name = name

    def critical(self,arg):
        self.logger.critical(arg)
        return True

    def debug(self,arg):
        self.logger.debug(arg)
        return True

    def error(self,arg):
        self.logger.error(arg)
        return True

    def info(self,arg):
        self.logger.info(arg)
        return True

    def mark(self,_):
        return '[%s]> %s' % (task_id(),_)

    def warning(self,arg):
        self.logger.warning(arg)
        return True
